@@ -1,6 +1,6 @@
 /* 
    HTTP request/response handling
-   Copyright (C) 1999-2006, Joe Orton <joe@manyfish.co.uk>
+   Copyright (C) 1999-2008, Joe Orton <joe@manyfish.co.uk>
 
    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Library General Public
@@ -27,13 +27,6 @@
 
 #include <sys/types.h>
 
-#ifdef HAVE_SYS_LIMITS_H
-#include <sys/limits.h>
-#endif
-#ifdef HAVE_LIMITS_H
-#include <limits.h> /* for UINT_MAX etc */
-#endif
-#include <ctype.h>
 #include <errno.h>
 #include <fcntl.h>
 #ifdef HAVE_STRING_H
@@ -49,7 +42,7 @@
 #include <unistd.h>
 #endif
 
-#include "ne_i18n.h"
+#include "ne_internal.h"
 
 #include "ne_alloc.h"
 #include "ne_request.h"
@@ -68,49 +61,11 @@ if (sret < 0) return aborted(req, msg, sret); } while (0)
 struct body_reader {
     ne_block_reader handler;
     ne_accept_response accept_response;
-    unsigned int use:1;
+    unsigned int use;
     void *userdata;
     struct body_reader *next;
 };
 
-#if !defined(LONG_LONG_MAX) && defined(LLONG_MAX)
-#define LONG_LONG_MAX LLONG_MAX
-#elif !defined(LONG_LONG_MAX) && defined(LONGLONG_MAX)
-#define LONG_LONG_MAX LONGLONG_MAX
-#endif
-
-#ifdef NE_LFS
-#define ne_lseek lseek64
-typedef off64_t ne_off_t;
-#define FMT_NE_OFF_T NE_FMT_OFF64_T
-#define NE_OFFT_MAX LONG_LONG_MAX
-#ifdef HAVE_STRTOLL
-#define ne_strtoff strtoll
-#else
-#define ne_strtoff strtoq
-#endif
-#else /* !NE_LFS */
-
-typedef off_t ne_off_t;
-#define ne_lseek lseek
-#define FMT_NE_OFF_T NE_FMT_OFF_T
-
-#if defined(SIZEOF_LONG_LONG) && defined(LONG_LONG_MAX) \
-    && SIZEOF_OFF_T == SIZEOF_LONG_LONG
-#define NE_OFFT_MAX LONG_LONG_MAX
-#else
-#define NE_OFFT_MAX LONG_MAX
-#endif
-
-#if SIZEOF_OFF_T > SIZEOF_LONG && defined(HAVE_STRTOLL)
-#define ne_strtoff strtoll
-#elif SIZEOF_OFF_T > SIZEOF_LONG && defined(HAVE_STRTOQ)
-#define ne_strtoff strtoq
-#else
-#define ne_strtoff strtol
-#endif
-#endif /* NE_LFS */
-
 struct field {
     char *name, *value;
     size_t vlen;
@@ -184,7 +139,7 @@ struct ne_request_s {
         ne_off_t progress; /* number of bytes read of response */
     } resp;
     
-    struct hook *private, *pre_send_hooks;
+    struct hook *private;
 
     /* response header fields */
     struct field *response_headers[HH_HASHSIZE];
@@ -195,15 +150,16 @@ struct ne_request_s {
     struct body_reader *body_readers;
 
     /*** Miscellaneous ***/
-    unsigned int method_is_head:1;
-    unsigned int use_expect100:1;
-    unsigned int can_persist:1;
+    unsigned int method_is_head;
+    unsigned int can_persist;
+
+    int flags[NE_REQFLAG_LAST];
 
     ne_session *session;
     ne_status status;
 };
 
-static int open_connection(ne_request *req);
+static int open_connection(ne_session *sess);
 
 /* Returns hash value for header 'name', converting it to lower-case
  * in-place. */
@@ -213,7 +169,7 @@ static inline unsigned int hash_and_lower(char *name)
     unsigned int hash = 0;
 
     for (pnt = name; *pnt != '\0'; pnt++) {
-	*pnt = tolower(*pnt);
+	*pnt = ne_tolower(*pnt);
 	hash = HH_ITERATE(hash,*pnt);
     }
 
@@ -234,15 +190,15 @@ static int aborted(ne_request *req, const char *doing, ssize_t code)
     switch(code) {
     case NE_SOCK_CLOSED:
 	if (sess->use_proxy) {
-	    ne_set_error(sess, _("%s: connection was closed by proxy server."),
+	    ne_set_error(sess, _("%s: connection was closed by proxy server"),
 			 doing);
 	} else {
-	    ne_set_error(sess, _("%s: connection was closed by server."),
+	    ne_set_error(sess, _("%s: connection was closed by server"),
 			 doing);
 	}
 	break;
     case NE_SOCK_TIMEOUT:
-	ne_set_error(sess, _("%s: connection timed out."), doing);
+	ne_set_error(sess, _("%s: connection timed out"), doing);
 	ret = NE_TIMEOUT;
 	break;
     case NE_SOCK_ERROR:
@@ -259,11 +215,10 @@ static int aborted(ne_request *req, const char *doing, ssize_t code)
     return ret;
 }
 
-static void notify_status(ne_session *sess, ne_conn_status status,
-			  const char *info)
+static void notify_status(ne_session *sess, ne_session_status status)
 {
     if (sess->notify_cb) {
-	sess->notify_cb(sess->notify_ud, status, info);
+	sess->notify_cb(sess->notify_ud, status, &sess->status);
     }
 }
 
@@ -285,72 +240,24 @@ void *ne_get_session_private(ne_session *sess, const char *id)
     return get_private(sess->private, id);
 }
 
-typedef void (*void_fn)(void);
-
-#define ADD_HOOK(hooks, fn, ud) add_hook(&(hooks), NULL, (void_fn)(fn), (ud))
-
-static void add_hook(struct hook **hooks, const char *id, void_fn fn, void *ud)
+void ne_set_request_private(ne_request *req, const char *id, void *userdata)
 {
     struct hook *hk = ne_malloc(sizeof (struct hook)), *pos;
 
-    if (*hooks != NULL) {
-	for (pos = *hooks; pos->next != NULL; pos = pos->next)
+    if (req->private != NULL) {
+	for (pos = req->private; pos->next != NULL; pos = pos->next)
 	    /* nullop */;
 	pos->next = hk;
     } else {
-	*hooks = hk;
+	req->private = hk;
     }
 
     hk->id = id;
-    hk->fn = fn;
-    hk->userdata = ud;
+    hk->fn = NULL;
+    hk->userdata = userdata;
     hk->next = NULL;
 }
 
-void ne_hook_create_request(ne_session *sess, 
-			    ne_create_request_fn fn, void *userdata)
-{
-    ADD_HOOK(sess->create_req_hooks, fn, userdata);
-}
-
-void ne_hook_pre_send(ne_session *sess, ne_pre_send_fn fn, void *userdata)
-{
-    ADD_HOOK(sess->pre_send_hooks, fn, userdata);
-}
-
-void ne_hook_post_send(ne_session *sess, ne_post_send_fn fn, void *userdata)
-{
-    ADD_HOOK(sess->post_send_hooks, fn, userdata);
-}
-
-void ne_hook_destroy_request(ne_session *sess,
-			     ne_destroy_req_fn fn, void *userdata)
-{
-    ADD_HOOK(sess->destroy_req_hooks, fn, userdata);    
-}
-
-void ne_hook_destroy_session(ne_session *sess,
-			     ne_destroy_sess_fn fn, void *userdata)
-{
-    ADD_HOOK(sess->destroy_sess_hooks, fn, userdata);
-}
-
-/* Hack to fix ne_compress layer problems */
-void ne__reqhook_pre_send(ne_request *req, ne_pre_send_fn fn, void *userdata)
-{
-    ADD_HOOK(req->pre_send_hooks, fn, userdata);
-}
-
-void ne_set_session_private(ne_session *sess, const char *id, void *userdata)
-{
-    add_hook(&sess->private, id, NULL, userdata);
-}
-
-void ne_set_request_private(ne_request *req, const char *id, void *userdata)
-{
-    add_hook(&req->private, id, NULL, userdata);
-}
-
 static ssize_t body_string_send(void *userdata, char *buffer, size_t count)
 {
     ne_request *req = userdata;
@@ -390,18 +297,20 @@ static ssize_t body_fd_send(void *userdata, char *buffer, size_t count)
             req->body.file.remain = req->body.file.length;
             return 0;
         } else {
-            char err[200];
+            char err[200], offstr[20];
 
             if (newoff == -1) {
                 /* errno was set */
                 ne_strerror(errno, err, sizeof err);
             } else {
                 strcpy(err, _("offset invalid"));
-            }                       
+            }
+            ne_snprintf(offstr, sizeof offstr, "%" FMT_NE_OFF_T,
+                        req->body.file.offset);
             ne_set_error(req->session, 
-                         _("Could not seek to offset %" FMT_NE_OFF_T 
+                         _("Could not seek to offset %s"
                            " of request body file: %s"), 
-                           req->body.file.offset, err);
+                           offstr, err);
             return -1;
         }
     }
@@ -429,11 +338,14 @@ static ssize_t body_fd_send(void *userdata, char *buffer, size_t count)
 static int send_request_body(ne_request *req, int retry)
 {
     ne_session *const sess = req->session;
-    ne_off_t progress = 0;
     char buffer[NE_BUFSIZ];
     ssize_t bytes;
 
     NE_DEBUG(NE_DBG_HTTP, "Sending request body:\n");
+
+    req->session->status.sr.progress = 0;
+    req->session->status.sr.total = req->body_length;
+    notify_status(sess, ne_status_sending);
     
     /* tell the source to start again from the beginning. */
     if (req->body_cb(req->body_ud, NULL, 0) != 0) {
@@ -453,12 +365,8 @@ static int send_request_body(ne_request *req, int retry)
 		 bytes, (int)bytes, buffer);
 
         /* invoke progress callback */
-        if (sess->progress_cb) {
-            progress += bytes;
-            /* TODO: progress_cb offset type mismatch ick */
-            req->session->progress_cb(sess->progress_ud, progress,
-                                      req->body_length);
-        }
+        req->session->status.sr.progress += bytes;
+        notify_status(sess, ne_status_sending);
     }
 
     if (bytes == 0) {
@@ -483,7 +391,7 @@ static void add_fixed_headers(ne_request *req)
      * close; otherwise, send Connection: Keep-Alive to pre-1.1 origin
      * servers to try harder to get a persistent connection, except if
      * using a proxy as per 2068§19.7.1.  Always add TE: trailers. */
-    if (req->session->no_persist) {
+    if (!req->session->flags[NE_SESSFLAG_PERSIST]) {
        ne_buffer_czappend(req->headers,
                           "Connection: TE, close" EOL
                           "TE: trailers" EOL);
@@ -516,6 +424,9 @@ ne_request *ne_request_create(ne_session *sess,
 
     req->session = sess;
     req->headers = ne_buffer_create();
+    
+    /* Presume the method is idempotent by default. */
+    req->flags[NE_REQFLAG_IDEMPOTENT] = 1;
 
     /* Add in the fixed headers */
     add_fixed_headers(req);
@@ -561,7 +472,7 @@ void ne_set_request_body_buffer(ne_request *req, const char *buffer,
     set_body_length(req, size);
 }
 
-void ne_set_request_body_provider(ne_request *req, off_t bodysize,
+void ne_set_request_body_provider(ne_request *req, ne_off_t bodysize,
 				  ne_provide_body provider, void *ud)
 {
     req->body_cb = provider;
@@ -570,7 +481,7 @@ void ne_set_request_body_provider(ne_request *req, off_t bodysize,
 }
 
 void ne_set_request_body_fd(ne_request *req, int fd,
-                            off_t offset, off_t length)
+                            ne_off_t offset, ne_off_t length)
 {
     req->body.file.fd = fd;
     req->body.file.offset = offset;
@@ -580,30 +491,19 @@ void ne_set_request_body_fd(ne_request *req, int fd,
     set_body_length(req, length);
 }
 
-#ifdef NE_LFS
-void ne_set_request_body_fd64(ne_request *req, int fd,
-                              off64_t offset, off64_t length)
+void ne_set_request_flag(ne_request *req, ne_request_flag flag, int value)
 {
-    req->body.file.fd = fd;
-    req->body.file.offset = offset;
-    req->body.file.length = length;
-    req->body_cb = body_fd_send;
-    req->body_ud = req;
-    set_body_length(req, length);
+    if (flag < NE_SESSFLAG_LAST) {
+        req->flags[flag] = value;
+    }
 }
 
-void ne_set_request_body_provider64(ne_request *req, off64_t bodysize,
-                                    ne_provide_body provider, void *ud)
+int ne_get_request_flag(ne_request *req, ne_request_flag flag)
 {
-    req->body_cb = provider;
-    req->body_ud = ud;
-    set_body_length(req, bodysize);
-}
-#endif
-
-void ne_set_request_expect100(ne_request *req, int flag)
-{
-    req->use_expect100 = flag;
+    if (flag < NE_REQFLAG_LAST) {
+        return req->flags[flag];
+    }
+    return -1;
 }
 
 void ne_add_request_header(ne_request *req, const char *name, 
@@ -744,8 +644,9 @@ void ne_request_destroy(ne_request *req)
     ne_buffer_destroy(req->headers);
 
     NE_DEBUG(NE_DBG_HTTP, "Running destroy hooks.\n");
-    for (hk = req->session->destroy_req_hooks; hk; hk = hk->next) {
+    for (hk = req->session->destroy_req_hooks; hk; hk = next_hk) {
 	ne_destroy_req_fn fn = (ne_destroy_req_fn)hk->fn;
+        next_hk = hk->next;
 	fn(req, hk->userdata);
     }
 
@@ -753,10 +654,6 @@ void ne_request_destroy(ne_request *req)
 	next_hk = hk->next;
 	ne_free(hk);
     }
-    for (hk = req->pre_send_hooks; hk; hk = next_hk) {
-	next_hk = hk->next;
-	ne_free(hk);
-    }
 
     if (req->status.reason_phrase)
 	ne_free(req->status.reason_phrase);
@@ -873,9 +770,9 @@ ssize_t ne_read_response_block(ne_request *req, char *buffer, size_t buflen)
     if (read_response_block(req, resp, buffer, &readlen))
 	return -1;
 
-    if (req->session->progress_cb) {
-	req->session->progress_cb(req->session->progress_ud, resp->progress, 
-				  resp->mode==R_CLENGTH ? resp->body.clen.total:-1);
+    if (readlen) {
+        req->session->status.sr.progress += readlen;
+        notify_status(req->session, ne_status_recving);
     }
 
     for (rdr = req->body_readers; rdr!=NULL; rdr=rdr->next) {
@@ -901,21 +798,17 @@ static ne_buffer *build_request(ne_request *req)
     /* Add custom headers: */
     ne_buffer_append(buf, req->headers->data, ne_buffer_size(req->headers));
 
-#define E100 "Expect: 100-continue" EOL
-    if (req->use_expect100)
-	ne_buffer_append(buf, E100, strlen(E100));
+    if (req->flags[NE_REQFLAG_EXPECT100]) {
+        ne_buffer_czappend(buf, "Expect: 100-continue\r\n");
+    }
 
     NE_DEBUG(NE_DBG_HTTP, "Running pre_send hooks\n");
     for (hk = req->session->pre_send_hooks; hk!=NULL; hk = hk->next) {
 	ne_pre_send_fn fn = (ne_pre_send_fn)hk->fn;
 	fn(req, hk->userdata, buf);
     }
-    for (hk = req->pre_send_hooks; hk!=NULL; hk = hk->next) {
-	ne_pre_send_fn fn = (ne_pre_send_fn)hk->fn;
-	fn(req, hk->userdata, buf);
-    }
     
-    ne_buffer_append(buf, "\r\n", 2);
+    ne_buffer_czappend(buf, "\r\n");
     return buf;
 }
 
@@ -975,10 +868,10 @@ static int read_status_line(ne_request *req, ne_status *status, int retry)
     if (status->reason_phrase) ne_free(status->reason_phrase);
     memset(status, 0, sizeof *status);
 
-    if (strncmp(buffer, "ICY ", 4) == 0 && strlen(buffer) > 8
+    /* Hack to allow ShoutCast-style servers, if requested. */
+    if (req->session->flags[NE_SESSFLAG_ICYPROTO]
+        && strncmp(buffer, "ICY ", 4) == 0 && strlen(buffer) > 8
         && buffer[7] == ' ') {
-        /* Hack for ShoutCast support; will be conditional in future
-         * releases. */
         status->code = atoi(buffer + 4);
         status->major_version = 1;
         status->minor_version = 0;
@@ -987,7 +880,7 @@ static int read_status_line(ne_request *req, ne_status *status, int retry)
         NE_DEBUG(NE_DBG_HTTP, "[status-line] ICY protocol; code %d\n", 
                  status->code);
     } else if (ne_parse_statusline(buffer, status)) {
-	return aborted(req, _("Could not parse response status line."), 0);
+	return aborted(req, _("Could not parse response status line"), 0);
     }
 
     return 0;
@@ -1024,7 +917,7 @@ static int send_request(ne_request *req, const ne_buffer *request)
     /* Send the Request-Line and headers */
     NE_DEBUG(NE_DBG_HTTP, "Sending request-line and headers:\n");
     /* Open the connection if necessary */
-    ret = open_connection(req);
+    ret = open_connection(sess);
     if (ret) return ret;
 
     /* Allow retry if a persistent connection has been used. */
@@ -1033,11 +926,11 @@ static int send_request(ne_request *req, const ne_buffer *request)
     sret = ne_sock_fullwrite(req->session->socket, request->data, 
                              ne_buffer_size(request));
     if (sret < 0) {
-	int aret = aborted(req, _("Could not send request"), ret);
+	int aret = aborted(req, _("Could not send request"), sret);
 	return RETRY_RET(retry, sret, aret);
     }
     
-    if (!req->use_expect100 && req->body_length > 0) {
+    if (!req->flags[NE_REQFLAG_EXPECT100] && req->body_length > 0) {
 	/* Send request body, if not using 100-continue. */
 	ret = send_request_body(req, retry);
 	if (ret) {
@@ -1056,7 +949,7 @@ static int send_request(ne_request *req, const ne_buffer *request)
 	/* Discard headers with the interim response. */
 	if ((ret = discard_headers(req)) != NE_OK) break;
 
-	if (req->use_expect100 && (status->code == 100)
+	if (req->flags[NE_REQFLAG_EXPECT100] && (status->code == 100)
             && req->body_length > 0 && !sentbody) {
 	    /* Send the body after receiving the first 100 Continue */
 	    if ((ret = send_request_body(req, 0)) != NE_OK) break;	    
@@ -1183,7 +1076,7 @@ static int read_response_headers(ne_request *req)
 	/* Convert the header name to lower case and hash it. */
 	for (pnt = hdr; (*pnt != '\0' && *pnt != ':' && 
 			 *pnt != ' ' && *pnt != '\t'); pnt++) {
-	    *pnt = tolower(*pnt);
+	    *pnt = ne_tolower(*pnt);
 	    hash = HH_ITERATE(hash,*pnt);
 	}
 
@@ -1209,7 +1102,7 @@ static int read_response_headers(ne_request *req)
 
     if (count == MAX_HEADER_FIELDS)
 	ret = aborted(
-	    req, _("Response exceeded maximum number of header fields."), 0);
+	    req, _("Response exceeded maximum number of header fields"), 0);
 
     return ret;
 }
@@ -1221,8 +1114,8 @@ static int lookup_host(ne_session *sess, struct host_info *info)
     if (sess->addrlist) return NE_OK;
 
     NE_DEBUG(NE_DBG_HTTP, "Doing DNS lookup on %s...\n", info->hostname);
-    if (sess->notify_cb)
-	sess->notify_cb(sess->notify_ud, ne_conn_namelookup, info->hostname);
+    sess->status.lu.hostname = info->hostname;
+    notify_status(sess, ne_status_lookup);
     info->address = ne_addr_resolve(info->hostname, 0);
     if (ne_addr_result(info->address)) {
 	char buf[256];
@@ -1240,25 +1133,29 @@ static int lookup_host(ne_session *sess, struct host_info *info)
 int ne_begin_request(ne_request *req)
 {
     struct body_reader *rdr;
-    struct host_info *host;
     ne_buffer *data;
     const ne_status *const st = &req->status;
     const char *value;
+    struct hook *hk;
     int ret;
 
-    /* Resolve hostname if necessary. */
-    host = req->session->use_proxy?&req->session->proxy:&req->session->server;
-    if (host->address == NULL) {
-        ret = lookup_host(req->session, host);
-        if (ret) return ret;
-    }    
-    
+    /* If a non-idempotent request is sent on a persisted connection,
+     * then it is impossible to distinguish between a server failure
+     * and a connection timeout if an EOF/RST is received.  So don't
+     * do that. */
+    if (!req->flags[NE_REQFLAG_IDEMPOTENT] && req->session->persisted
+        && !req->session->flags[NE_SESSFLAG_CONNAUTH]) {
+        NE_DEBUG(NE_DBG_HTTP, "req: Closing connection for non-idempotent "
+                 "request.\n");
+        ne_close_connection(req->session);
+    }
+
     /* Build the request string, and send it */
     data = build_request(req);
     DEBUG_DUMP_REQUEST(data->data);
     ret = send_request(req, data);
     /* Retry this once after a persistent connection timeout. */
-    if (ret == NE_RETRY && !req->session->no_persist) {
+    if (ret == NE_RETRY) {
 	NE_DEBUG(NE_DBG_HTTP, "Persistent connection timed out, retrying.\n");
 	ret = send_request(req, data);
     }
@@ -1321,15 +1218,23 @@ int ne_begin_request(ne_request *req)
      * regardless of what headers are present. */
     if (req->method_is_head || st->code == 204 || st->code == 304) {
     	req->resp.mode = R_NO_BODY;
-    } else if (get_response_header_hv(req, HH_HV_TRANSFER_ENCODING,
-                                      "transfer-encoding")) {
-        /* Treat *any* t-e header as implying a chunked response
-         * regardless of value, per the "Protocol Compliance"
-         * statement in the manual. */
-        req->resp.mode = R_CHUNKED;
-        req->resp.body.chunk.remain = 0;
-    } else if ((value = get_response_header_hv(req, HH_HV_CONTENT_LENGTH,
-                                               "content-length")) != NULL) {
+    }
+    /* Broken intermediaries exist which use "transfer-encoding: identity"
+     * to mean "no transfer-coding".  So that case must be ignored. */
+    else if ((value = get_response_header_hv(req, HH_HV_TRANSFER_ENCODING,
+                                             "transfer-encoding")) != NULL
+             && ne_strcasecmp(value, "identity") != 0) {
+        /* Otherwise, fail iff an unknown transfer-coding is used. */
+        if (ne_strcasecmp(value, "chunked") == 0) {
+            req->resp.mode = R_CHUNKED;
+            req->resp.body.chunk.remain = 0;
+        }
+        else {
+            return aborted(req, _("Unknown transfer-coding in response"), 0);
+        }
+    } 
+    else if ((value = get_response_header_hv(req, HH_HV_CONTENT_LENGTH,
+                                             "content-length")) != NULL) {
         ne_off_t len = ne_strtoff(value, NULL, 10);
         if (len != NE_OFFT_MAX && len >= 0) {
             req->resp.mode = R_CLENGTH;
@@ -1337,17 +1242,28 @@ int ne_begin_request(ne_request *req)
         } else {
             /* fail for an invalid content-length header. */
             return aborted(req, _("Invalid Content-Length in response"), 0);
-        }            
+        }
     } else {
         req->resp.mode = R_TILLEOF; /* otherwise: read-till-eof mode */
     }
     
+    NE_DEBUG(NE_DBG_HTTP, "Running post_headers hooks\n");
+    for (hk = req->session->post_headers_hooks; hk != NULL; hk = hk->next) {
+        ne_post_headers_fn fn = (ne_post_headers_fn)hk->fn;
+        fn(req, hk->userdata, &req->status);
+    }
+    
     /* Prepare for reading the response entity-body.  Call each of the
      * body readers and ask them whether they want to accept this
      * response or not. */
     for (rdr = req->body_readers; rdr != NULL; rdr=rdr->next) {
 	rdr->use = rdr->accept_response(rdr->userdata, req, st);
     }
+
+    req->session->status.sr.progress = 0;
+    req->session->status.sr.total = 
+        req->resp.mode == R_CLENGTH ? req->resp.body.clen.total : -1;
+    notify_status(req->session, ne_status_recving);
     
     return NE_OK;
 }
@@ -1374,7 +1290,7 @@ int ne_end_request(ne_request *req)
     
     /* Close the connection if persistent connections are disabled or
      * not supported by the server. */
-    if (req->session->no_persist || !req->can_persist)
+    if (!req->session->flags[NE_SESSFLAG_PERSIST] || !req->can_persist)
 	ne_close_connection(req->session);
     else
 	req->session->persisted = 1;
@@ -1470,9 +1386,11 @@ static int proxy_tunnel(ne_session *sess)
     sess->persisted = 0; /* don't treat this is a persistent connection. */
 
     if (ret != NE_OK || !sess->connected || req->status.klass != 2) {
-	ne_set_error
-	    (sess, _("Could not create SSL connection through proxy server"));
-	ret = NE_ERROR;
+        char *err = ne_strdup(sess->error);
+        ne_set_error(sess, _("Could not create SSL connection "
+                             "through proxy server: %s"), err);
+        ne_free(err);
+        if (ret == NE_OK) ret = NE_ERROR;
     }
 
     ne_request_destroy(req);
@@ -1511,9 +1429,8 @@ static const ne_inet_addr *resolve_next(ne_session *sess,
  * that once a connection to a particular network address has
  * succeeded, that address will be used first for the next attempt to
  * connect. */
-static int do_connect(ne_request *req, struct host_info *host, const char *err)
+static int do_connect(ne_session *sess, struct host_info *host, const char *err)
 {
-    ne_session *const sess = req->session;
     int ret;
 
     if ((sess->socket = ne_sock_create()) == NULL) {
@@ -1521,11 +1438,20 @@ static int do_connect(ne_request *req, struct host_info *host, const char *err)
         return NE_ERROR;
     }
 
+    if (sess->cotimeout)
+	ne_sock_connect_timeout(sess->socket, sess->cotimeout);
+
+    if (sess->local_addr)
+        ne_sock_prebind(sess->socket, sess->local_addr, 0);
+
     if (host->current == NULL)
 	host->current = resolve_first(sess, host);
 
+    sess->status.ci.hostname = host->hostname;
+
     do {
-	notify_status(sess, ne_conn_connecting, host->hostport);
+        sess->status.ci.address = host->current;
+	notify_status(sess, ne_status_connecting);
 #ifdef NE_DEBUGGING
 	if (ne_debug_mask & NE_DBG_HTTP) {
 	    char buf[150];
@@ -1540,10 +1466,10 @@ static int do_connect(ne_request *req, struct host_info *host, const char *err)
     if (ret) {
         ne_set_error(sess, "%s: %s", err, ne_sock_error(sess->socket));
         ne_sock_close(sess->socket);
-	return NE_CONNECT;
+	return ret == NE_SOCK_TIMEOUT ? NE_TIMEOUT : NE_CONNECT;
     }
 
-    notify_status(sess, ne_conn_connected, host->hostport);
+    notify_status(sess, ne_status_connected);
     
     if (sess->rdtimeout)
 	ne_sock_read_timeout(sess->socket, sess->rdtimeout);
@@ -1554,30 +1480,35 @@ static int do_connect(ne_request *req, struct host_info *host, const char *err)
     return NE_OK;
 }
 
-static int open_connection(ne_request *req) 
+static int open_connection(ne_session *sess) 
 {
-    ne_session *sess = req->session;
     int ret;
+    struct host_info *host;
     
     if (sess->connected) return NE_OK;
 
-    if (!sess->use_proxy)
-	ret = do_connect(req, &sess->server, _("Could not connect to server"));
-    else
-	ret = do_connect(req, &sess->proxy,
-			 _("Could not connect to proxy server"));
-
+    /* Resolve hostname if necessary. */
+    host = sess->use_proxy ? &sess->proxy : &sess->server;
+    if (host->address == NULL) {
+        ret = lookup_host(sess, host);
+        if (ret) return ret;
+    }    
+    
+    ret = do_connect(sess, host, 
+                     sess->use_proxy ? 
+                     _("Could not connect to proxy server")
+                     : _("Could not connect to server"));
     if (ret != NE_OK) return ret;
 
 #ifdef NE_HAVE_SSL
     /* Negotiate SSL layer if required. */
     if (sess->use_ssl && !sess->in_connect) {
         /* CONNECT tunnel */
-        if (req->session->use_proxy)
+        if (sess->use_proxy)
             ret = proxy_tunnel(sess);
         
         if (ret == NE_OK) {
-            ret = ne__negotiate_ssl(req);
+            ret = ne__negotiate_ssl(sess);
             if (ret != NE_OK)
                 ne_close_connection(sess);
         }
