@@ -1,6 +1,6 @@
 /* 
    HTTP request handling tests
-   Copyright (C) 2001-2003, Joe Orton <joe@manyfish.co.uk>
+   Copyright (C) 2001-2008, Joe Orton <joe@manyfish.co.uk>
 
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
@@ -30,6 +30,7 @@
 #ifdef HAVE_UNISTD_H
 #include <unistd.h>
 #endif
+#include <fcntl.h>
 
 #include "ne_request.h"
 #include "ne_socket.h"
@@ -69,10 +70,11 @@ static int finish_request(void)
 #define TE_CHUNKED "Transfer-Encoding: chunked\r\n"
 
 /* takes response body chunks and appends them to a buffer. */
-static void collector(void *ud, const char *data, size_t len)
+static int collector(void *ud, const char *data, size_t len)
 {
     ne_buffer *buf = ud;
     ne_buffer_append(buf, data, len);
+    return 0;
 }
 
 typedef ne_request *(*construct_request)(ne_session *sess, void *userdata);
@@ -107,29 +109,32 @@ static int run_request(ne_session *sess, int status,
     return OK;
 }
 
-/* Runs a server function 'fn', expecting to get a header 'name' with value
- * 'value' in the response. */
+/* Runs a server function 'fn', expecting to get a header 'name' with
+ * value 'value' in the response.  If 'value' is NULL, expects that
+ * *no* header of that name is present. */
 static int expect_header_value(const char *name, const char *value,
 			       server_fn fn, void *userdata)
 {
     ne_session *sess;
     ne_request *req;
-    char *gotval = NULL;
+    const char *gotval;
 
     CALL(make_session(&sess, fn, userdata));
 
     req = ne_request_create(sess, "FOO", "/bar");
-    ne_add_response_header_handler(req, name, ne_duplicate_header, &gotval);
     ONREQ(ne_request_dispatch(req));
     CALL(await_server());
     
-    ONN("no header value set", gotval == NULL);
-    ONV(strcmp(gotval, value),
-	("header value mis-match: got [%s] not [%s]", gotval, value));
+    gotval = ne_get_response_header(req, name);
+    ONV(value && !gotval, ("header '%s: %s' not sent", name, value));
+    ONV(!value && gotval, ("header '%s: %s' not expected", name, gotval));
+
+    ONV(value && gotval && strcmp(gotval, value),
+	("header '%s' mis-match: got '%s' not '%s'", 
+         name, gotval, value));
     
     ne_request_destroy(req);
     ne_session_destroy(sess);
-    ne_free(gotval);
 
     return OK;
 }
@@ -198,7 +203,7 @@ static int reason_phrase(void)
 
     CALL(make_session(&sess, single_serve_string, RESP200
 		      "Connection: close\r\n\r\n"));
-    CALL(any_request(sess, "/foo"));
+    ONREQ(any_request(sess, "/foo"));
     CALL(await_server());
     
     ONV(strcmp(ne_get_error(sess), "200 OK"),
@@ -298,13 +303,15 @@ static int te_header(void)
 			   "\r\n" ABCDE_CHUNKS);
 }
 
-/* test that the presence of *any* t-e header implies a chunked
- * response. */
-static int any_te_header(void)
+static int te_identity(void)
 {
-    return expect_response("abcde", single_serve_string, RESP200
-                           "Transfer-Encoding: punked\r\n" "\r\n"
-                           ABCDE_CHUNKS);
+    /* http://bugzilla.gnome.org/show_bug.cgi?id=310636 says privoxy
+     * uses the "identity" transfer-coding. */
+    return expect_response("abcde", single_serve_string,
+			   RESP200 "Transfer-Encoding: identity\r\n"
+                           "Content-Length: 5\r\n"
+			   "\r\n"
+                           "abcde");
 }
 
 static int chunk_numeric(void)
@@ -723,54 +730,176 @@ static int continued_header(void)
 			       NO_BODY);
 }
 
-static void mh_header(void *ctx, const char *value)
-{
-    int *state = ctx;
-    static const char *hdrs[] = { "jim", "jab", "jar" };
-
-    if (*state < 0 || *state > 2) {
-	/* already failed. */
-	return;
-    }
-
-    if (strcmp(value, hdrs[*state]))
-	*state = -*state;
-    else
-	(*state)++;
-}
-
 /* check headers callbacks are working correctly. */
 static int multi_header(void)
 {
-    ne_session *sess = ne_session_create("http", "localhost", 7777);
+    return expect_header_value("X-Header", "jim, jab, jar",
+                               single_serve_string,
+                               RESP200 
+                               "X-Header: jim\r\n" 
+                               "x-header: jab\r\n"
+                               "x-Header: jar\r\n"
+                               "Content-Length: 0\r\n\r\n");
+}
+
+/* check headers callbacks are working correctly. */
+static int multi_header2(void)
+{
+    return expect_header_value("X-Header", "jim, jab, jar",
+                               single_serve_string,
+                               RESP200 
+                               "X-Header: jim  \r\n" 
+                               "x-header: jab  \r\n"
+                               "x-Header: jar  \r\n"
+                               "Content-Length: 0\r\n\r\n");
+}
+
+/* RFC 2616 14.10: headers listed in Connection must be stripped on
+ * receiving an HTTP/1.0 message in case there was a pre-1.1 proxy
+ * somewhere. */
+static int strip_http10_connhdr(void)
+{
+    return expect_header_value("X-Widget", NULL,
+                               single_serve_string,
+                               "HTTP/1.0 200 OK\r\n"
+                               "Connection: x-widget\r\n"
+                               "x-widget: blah\r\n"
+                               "Content-Length: 0\r\n"
+                               "\r\n");
+}
+
+static int strip_http10_connhdr2(void)
+{
+    return expect_header_value("X-Widget", NULL,
+                               single_serve_string,
+                               "HTTP/1.0 200 OK\r\n"
+                               "Connection: connection, x-fish, x-widget\r\n"
+                               "x-widget: blah\r\n"
+                               "Content-Length: 0\r\n"
+                               "\r\n");
+}
+
+static int post_send_retry(ne_request *req, void *userdata, 
+                           const ne_status *status)
+{
+    return status->code == 400 ? NE_RETRY : NE_OK;    
+}
+
+/* Test that the stored response headers are forgotten if the request
+ * is retried. */
+static int reset_headers(void)
+{
+    ne_session *sess;
     ne_request *req;
-    int state = 0;
+    const char *value;
 
-    ON(sess == NULL);
-    ON(spawn_server(7777, single_serve_string, 
-		    RESP200 
-		    "X-Header: jim\r\n" 
-		    "x-header: jab\r\n"
-		    "x-Header: jar\r\n"
-		    "Content-Length: 0\r\n\r\n"));
+    CALL(make_session(&sess, single_serve_string,
+                      "HTTP/1.1 400 Hit me again\r\n"
+                      "Content-Length: 0\r\n"
+                      "X-Foo: bar\r\n" "\r\n"
 
-    req = ne_request_create(sess, "GET", "/");
-    ON(req == NULL);
+                      "HTTP/1.1 200 Thank you kindly\r\n"
+                      "Content-Length: 0\r\n"
+                      "X-Foo: hello fair world\r\n" "\r\n"));
 
-    ne_add_response_header_handler(req, "x-header", mh_header, &state);
+    ne_hook_post_send(sess, post_send_retry, NULL);
+    
+    req = ne_request_create(sess, "GET", "/foo");
 
     ONREQ(ne_request_dispatch(req));
 
-    ON(await_server());
+    value = ne_get_response_header(req, "X-Foo");
+    ONCMP("hello fair world", value, "response header", "X-Foo");
 
-    ON(state != 3);
+    ne_request_destroy(req);
+    ne_session_destroy(sess);
+    CALL(await_server());
 
+    return OK;
+}
+
+static int iterate_none(void)
+{
+    ne_session *sess;
+    ne_request *req;
+
+    CALL(make_session(&sess, single_serve_string,
+                      "HTTP/1.0 200 OK\r\n\r\n"));
+
+    req = ne_request_create(sess, "GET", "/");
+    ONREQ(ne_request_dispatch(req));
+
+    ONN("iterator was not NULL for no headers",
+        ne_response_header_iterate(req, NULL, NULL, NULL) != NULL);
+
+    CALL(await_server());
     ne_request_destroy(req);
     ne_session_destroy(sess);
 
     return OK;
 }
 
+#define MANY_HEADERS (90)
+
+static int iterate_many(void)
+{
+    ne_request *req;
+    ne_buffer *buf = ne_buffer_create();
+    ne_session *sess;
+    int n;
+    struct header {
+        char name[10], value[10];
+        int seen;
+    } hdrs[MANY_HEADERS];
+    void *cursor = NULL;
+    const char *name, *value;
+    
+    ne_buffer_czappend(buf, "HTTP/1.0 200 OK\r\n");
+
+    for (n = 0; n < MANY_HEADERS; n++) {
+        sprintf(hdrs[n].name, "x-%d", n);
+        sprintf(hdrs[n].value, "Y-%d", n);
+        hdrs[n].seen = 0;
+        
+        ne_buffer_concat(buf, hdrs[n].name, ": ", hdrs[n].value, "\r\n", NULL);
+    }
+    
+    ne_buffer_czappend(buf, "\r\n");
+
+    CALL(make_session(&sess, single_serve_string, buf->data));
+
+    req = ne_request_create(sess, "GET", "/foo");
+    ONREQ(ne_request_dispatch(req));
+
+    while ((cursor = ne_response_header_iterate(req, cursor, &name, &value))) {
+        n = -1;
+
+        ONV(strncmp(name, "x-", 2) || strncmp(value, "Y-", 2)
+            || strcmp(name + 2, value + 2)
+            || (n = atoi(name + 2)) >= MANY_HEADERS
+            || n < 0,
+            ("bad name/value pair: %s = %s", name, value));
+
+        NE_DEBUG(NE_DBG_HTTP, "iterate: got pair (%d): %s = %s\n",
+                 n, name, value);
+
+        ONV(hdrs[n].seen == 1, ("duplicate pair %d", n));
+        hdrs[n].seen = 1;
+    }
+           
+    for (n = 0; n < MANY_HEADERS; n++) {
+        ONV(hdrs[n].seen == 0, ("unseen pair %d", n));
+    }
+    
+    ne_buffer_destroy(buf);
+    ne_request_destroy(req);
+    ne_session_destroy(sess);
+    CALL(await_server());
+    
+    return OK;
+}
+
+
 struct s1xx_args {
     int count;
     int hdrs;
@@ -824,17 +953,12 @@ static int skip_1xx_hdrs(void)
 
 #undef sess
 
-/* server for expect_100_once: eats a dummy request, then serves a
- * 100-continue request, and fails if the request body is sent
- * twice. */
+/* server for expect_100_once: serves a 100-continue request, and
+ * fails if the request body is sent twice. */
 static int serve_100_once(ne_socket *sock, void *ud)
 {
     struct s1xx_args args = {2, 0};
     char ch;
-    /* dummy first request. */
-    CALL(discard_request(sock));
-    CALL(SEND_STRING(sock, RESP200 "Content-Length: 0\r\n\r\n"));
-    /* now the real 1xx request. */
     CALL(serve_1xx(sock, &args));
     CALL(discard_body(sock));
     ONN("body was served twice", ne_sock_read(sock, &ch, 1) == 1);
@@ -851,24 +975,37 @@ static int expect_100_once(void)
     char body[BUFSIZ];
 
     CALL(make_session(&sess, serve_100_once, NULL));
-    ne_set_expect100(sess, 1);
 
-    /* 100-continue is only used if the server is known to claim
-     * HTTP/1.1 compliance; make a dummy request on the socket first,
-     * to trigger that logic. */
-    CALL(any_request(sess, "/foo"));
-    
-    /* now the real request. */
     req = ne_request_create(sess, "GET", "/foo");
+    ne_set_request_flag(req, NE_REQFLAG_EXPECT100, 1);
+    ONN("expect100 flag ignored",
+        ne_get_request_flag(req, NE_REQFLAG_EXPECT100) != 1);
     memset(body, 'A', sizeof(body));
     ne_set_request_body_buffer(req, body, sizeof(body));
-    ONN("request failed", ne_request_dispatch(req));
+    ONREQ(ne_request_dispatch(req));
     ne_request_destroy(req);
     ne_session_destroy(sess);
     CALL(await_server());
     return OK;
 }
 
+/* regression test for enabling 100-continue without sending a body. */
+static int expect_100_nobody(void)
+{
+    ne_session *sess;
+    ne_request *req;
+
+    CALL(make_session(&sess, serve_100_once, NULL));
+    
+    req = ne_request_create(sess, "GET", "/foo");
+    ne_set_request_flag(req, NE_REQFLAG_EXPECT100, 1);
+    ONREQ(ne_request_dispatch(req));
+    ne_request_destroy(req);
+    ne_session_destroy(sess);
+
+    return await_server();
+}
+
 struct body {
     char *body;
     size_t size;
@@ -965,19 +1102,6 @@ static int send_bodies(void)
     return OK;
 }
 
-static int serve_infinite_headers(ne_socket *sock, void *userdata)
-{
-    CALL(discard_request(sock));
-
-    SEND_STRING(sock, RESP200);
-    
-    for (;;) {
-	SEND_STRING(sock, "x-foo: bar\r\n");
-    }
-
-    return 0;	
-}
-
 /* Utility function: run a request using the given server fn, and the
  * request should fail. If 'error' is non-NULL, it must be a substring
  * of the error string. */
@@ -1046,7 +1170,8 @@ static int fail_request(int with_body, server_fn fn, void *ud, int forever)
 
 static int unbounded_headers(void)
 {
-    return fail_request(0, serve_infinite_headers, NULL, 0);
+    struct infinite i = { RESP200, "x-foo: bar\r\n" };
+    return fail_request(0, serve_infinite, &i, 0);
 }
 
 static int blank_response(void)
@@ -1068,18 +1193,11 @@ static int not_http(void)
     return fail_request(0, serve_non_http, NULL, 0);
 }
 
-static int serve_infinite_folds(ne_socket *sock, void *ud)
-{
-    SEND_STRING(sock, "HTTP/1.0 200 OK\r\nFoo: bar\r\n");
-    for (;;) {
-	SEND_STRING(sock, "  hello there.\r\n");
-    }
-    return OK;
-}
-
 static int unbounded_folding(void)
 {
-    return fail_request(0, serve_infinite_folds, NULL, 0);
+    struct infinite i = { "HTTP/1.0 200 OK\r\nFoo: bar\r\n", 
+                          "  hello there.\r\n" };
+    return fail_request(0, serve_infinite, &i, 0);
 }
 
 static int serve_close(ne_socket *sock, void *ud)
@@ -1158,13 +1276,15 @@ static enum {
     prog_done /* finished. */
 } prog_state = prog_transfer;
 
-static off_t prog_last = -1, prog_total;
+static ne_off_t prog_last = -1, prog_total;
+
+#define FOFF "%" NE_FMT_NE_OFF_T
 
 /* callback for send_progress. */
-static void s_progress(void *userdata, off_t prog, off_t total)
+static void s_progress(void *userdata, ne_off_t prog, ne_off_t total)
 {
     NE_DEBUG(NE_DBG_HTTP, 
-	     "progress callback: %" NE_FMT_OFF_T "/%" NE_FMT_OFF_T ".\n",
+	     "progress callback: " FOFF "/" FOFF ".\n",
 	     prog, total);
 
     switch (prog_state) {
@@ -1173,19 +1293,19 @@ static void s_progress(void *userdata, off_t prog, off_t total)
 	return;
     case prog_transfer:
 	if (total != prog_total) {
-	    t_context("total unexpected: %ld not %ld", total, prog_total);
+	    t_context("total unexpected: " FOFF " not " FOFF "", total, prog_total);
 	    prog_state = prog_error;
 	}
 	else if (prog > total) {
-	    t_context("first progress was invalid (%ld/%ld)", prog, total);
+	    t_context("first progress was invalid (" FOFF "/" FOFF ")", prog, total);
 	    prog_state = prog_error;
 	}
 	else if (prog_last != -1 && prog_last > prog) {
-	    t_context("progess went backwards: %ld to %ld", prog_last, prog);
+	    t_context("progess went backwards: " FOFF " to " FOFF, prog_last, prog);
 	    prog_state = prog_error;
 	}
 	else if (prog_last == prog) {
-	    t_context("no progress made! %ld to %ld", prog_last, prog);
+	    t_context("no progress made! " FOFF " to " FOFF, prog_last, prog);
 	    prog_state = prog_error;
 	}
 	else if (prog == total) {
@@ -1197,6 +1317,8 @@ static void s_progress(void *userdata, off_t prog, off_t total)
     prog_last = prog;
 }
 
+#undef FOFF
+
 static ssize_t provide_progress(void *userdata, char *buf, size_t bufsiz)
 {
     int *count = userdata;
@@ -1289,7 +1411,7 @@ static int fail_lookup(void)
  * requests on the session would crash. */
 static int fail_double_lookup(void)
 {
-     ne_session *sess = ne_session_create("http", "nohost.example.com", 80);
+     ne_session *sess = ne_session_create("http", "nonesuch.invalid", 80);
      ONN("request did not give lookup failure",
 	 any_request(sess, "/foo") != NE_LOOKUP);
      ONN("second request did not give lookup failure",
@@ -1307,7 +1429,7 @@ static int fail_connect(void)
  * request. */
 static int proxy_no_resolve(void)
 {
-     ne_session *sess = ne_session_create("http", "no.such.domain", 80);
+     ne_session *sess = ne_session_create("http", "nonesuch2.invalid", 80);
      int ret;
      
      ne_session_proxy(sess, "localhost", 7777);
@@ -1409,8 +1531,8 @@ static int fail_statusline(void)
     ret = any_request(sess, "/fail");
     ONV(ret != NE_ERROR, ("request failed with %d not NE_ERROR", ret));
     
-    /* FIXME: will break for i18n. */
-    ONV(strcmp(ne_get_error(sess), "Could not parse response status line."),
+    ONV(strstr(ne_get_error(sess), 
+               "Could not parse response status line") == NULL,
 	("session error was `%s'", ne_get_error(sess)));
 
     ne_session_destroy(sess);
@@ -1433,20 +1555,37 @@ static int fail_long_header(void)
     return invalid_response_gives_error(resp, "Line too long");
 }
 
-static int fail_corrupt_chunks(void)
+static int fail_on_invalid(void)
 {
     static const struct {
         const char *resp, *error;
     } ts[] = {
-        /* not CRLF */
+        /* non-chunked TE. */
+        { RESP200 "transfer-encoding: punked\r\n" "\r\n" ABCDE_CHUNKS ,
+          "Unknown transfer-coding" },
+        /* chunk without trailing CRLF */
         { RESP200 TE_CHUNKED "\r\n" "5\r\n" "abcdeFISH", 
           "delimiter was invalid" },
-        /* short CRLF */
+        /* chunk with CR then EOF */
         { RESP200 TE_CHUNKED "\r\n" "5\r\n" "abcde\n",
           "not read chunk delimiter" },
-        /* CR-notLF */
+        /* chunk with CR then notLF */
         { RESP200 TE_CHUNKED "\r\n" "5\r\n" "abcde\rZZZ",
           "delimiter was invalid" },
+        /* chunk size overflow */
+        { RESP200 TE_CHUNKED "\r\n" "800000000\r\n" "abcde\r\n",
+          "Could not parse chunk size" },
+        /* EOF at chunk size */
+        { RESP200 TE_CHUNKED "\r\n", "Could not read chunk size" },
+
+        /* negative C-L */
+        { RESP200 "Content-Length: -1\r\n" "\r\n" "abcde",
+          "Invalid Content-Length" },
+        /* stupidly-large C-L */
+        { RESP200 "Content-Length: 99999999999999999999999999\r\n" 
+          "\r\n" "abcde",
+          "Invalid Content-Length" },
+        
         { NULL, NULL }
     };
     int n;
@@ -1468,12 +1607,12 @@ static int versions(void)
 		      "HTTP/1.0 200 OK\r\n"
 		      "Content-Length: 0\r\n\r\n"));
     
-    CALL(any_request(sess, "/http11"));
+    ONREQ(any_request(sess, "/http11"));
 	 
     ONN("did not detect HTTP/1.1 compliance",
 	ne_version_pre_http11(sess) != 0);
 	 
-    CALL(any_request(sess, "/http10"));
+    ONREQ(any_request(sess, "/http10"));
 
     ONN("did not detect lack of HTTP/1.1 compliance",
 	ne_version_pre_http11(sess) == 0);
@@ -1516,7 +1655,7 @@ static int hook_create_req(void)
     args.uri = "/foo";
     args.result = -1;
 
-    CALL(any_request(sess, "/foo"));
+    ONREQ(any_request(sess, "/foo"));
     
     ONN("first hook never called", args.result == -1);
     if (args.result) return FAIL;
@@ -1527,7 +1666,7 @@ static int hook_create_req(void)
     /* force use of absoluteURI in request-uri */
     ne_session_proxy(sess, "localhost", 7777);
 
-    CALL(any_request(sess, "/bar"));
+    ONREQ(any_request(sess, "/bar"));
     
     ONN("second hook never called", args.result == -1);
     if (args.result) return FAIL;
@@ -1574,6 +1713,392 @@ static int dup_method(void)
     return OK;
 }
 
+static int abortive_reader(void *userdata, const char *buf, size_t len)
+{
+    ne_session *sess = userdata;
+    if (len == 5 && strncmp(buf, "abcde", 5) == 0) {
+        ne_set_error(sess, "Reader callback failed");
+    } else {
+        ne_set_error(sess, "Reader callback called with length %" NE_FMT_SIZE_T,
+                     len);
+    }
+    return NE_ERROR;
+}
+
+static int abort_reader(void)
+{
+    ne_session *sess;
+    ne_request *req;
+    int ret;
+
+    CALL(make_session(&sess, single_serve_string, 
+                      RESP200 "Content-Length: 5\r\n\r\n"
+                      "abcde"
+                      "HTTP/1.1 200 OK\r\n"
+                      "Content-Length: 0\r\n\r\n"));
+
+    req = ne_request_create(sess, "GET", "/foo");
+    ne_add_response_body_reader(req, ne_accept_2xx, abortive_reader, sess);
+    ret = ne_request_dispatch(req);
+    ONV(ret != NE_ERROR, ("request did not fail with NE_ERROR: %d", ret));
+    ONV(strcmp(ne_get_error(sess), "Reader callback failed") != 0,
+        ("unexpected session error string: %s", ne_get_error(sess)));
+    ne_request_destroy(req);
+    /* test that the connection was closed. */
+    ONN("connection not closed after aborted response",
+        any_2xx_request(sess, "/failmeplease") == OK);
+    ne_session_destroy(sess);
+    CALL(await_server());
+    return OK;
+}
+
+/* attempt and fail to send request from offset 500 of /dev/null. */
+static int send_bad_offset(void)
+{
+    ne_session *sess;
+    ne_request *req;
+    int ret, fds[2];
+
+    CALL(make_session(&sess, single_serve_string,
+                      RESP200 "Content-Length: 0\r\n" "\r\n"));
+
+    /* create a pipe, on which seek is guaranteed to fail. */
+    ONN("could not create pipe", pipe(fds) != 0);
+
+    req = ne_request_create(sess, "PUT", "/null");
+
+    ne_set_request_body_fd(req, fds[0], 500, 5);
+    
+    ret = ne_request_dispatch(req);
+
+    close(fds[0]);
+    close(fds[1]);
+
+    ONN("request dispatched with bad offset!", ret == NE_OK);
+    ONV(ret != NE_ERROR,
+        ("request failed with non-NE_ERROR: %s", ne_get_error(sess)));
+
+    ONV(strstr(ne_get_error(sess), "Could not seek") == NULL,
+        ("bad error message from seek failure: %s", ne_get_error(sess)));
+
+    reap_server();
+    ne_request_destroy(req);
+    ne_session_destroy(sess);
+    return OK;
+}
+
+static void thook_create_req(ne_request *req, void *userdata,
+                             const char *method, const char *requri)
+{
+    ne_buffer *buf = userdata;
+
+    ne_buffer_concat(buf, "(create,", method, ",", requri, ")\n", NULL);
+}
+
+static void hook_pre_send(ne_request *req, void *userdata, 
+                          ne_buffer *header)
+{
+    ne_buffer *buf = userdata;
+
+    ne_buffer_czappend(buf, "(pre-send)\n");
+}
+
+/* Returns a static string giving a comma-separated representation of
+ * the status structure passed in. */ 
+static char *status_to_string(const ne_status *status)
+{
+    static char sbuf[128];
+
+    ne_snprintf(sbuf, sizeof sbuf, "HTTP/%d.%d,%d,%s", 
+                status->major_version, status->minor_version,
+                status->code, status->reason_phrase);
+
+    return sbuf;
+}
+
+static void hook_post_headers(ne_request *req, void *userdata,
+			      const ne_status *status)
+{
+    ne_buffer *buf = userdata;
+
+    ne_buffer_concat(buf, "(post-headers,", status_to_string(status), ")\n",
+		     NULL);
+}
+
+
+static int hook_post_send(ne_request *req, void *userdata,
+                          const ne_status *status)
+{
+    ne_buffer *buf = userdata;
+
+    ne_buffer_concat(buf, "(post-send,", status_to_string(status), ")\n", 
+		     NULL);
+
+    return NE_OK;
+}
+
+static void hook_destroy_req(ne_request *req, void *userdata)
+{
+    ne_buffer *buf = userdata;
+
+    ne_buffer_czappend(buf, "(destroy-req)\n");
+}
+
+static void hook_destroy_sess(void *userdata)
+{
+    ne_buffer *buf = userdata;
+
+    ne_buffer_czappend(buf, "(destroy-sess)\n");
+}
+
+static void hook_close_conn(void *userdata)
+{
+    ne_buffer *buf = userdata;
+
+    ne_buffer_czappend(buf, "(close-conn)\n");
+}
+
+static int hooks(void)
+{
+    ne_buffer *buf = ne_buffer_create();
+    ne_session *sess;
+    struct many_serve_args args;
+
+    args.str = RESP200 "Content-Length: 0\r\n" "\r\n";
+    args.count = 3;
+
+    CALL(make_session(&sess, many_serve_string, &args));
+
+    ne_hook_create_request(sess, thook_create_req, buf);
+    ne_hook_pre_send(sess, hook_pre_send, buf);
+    ne_hook_post_headers(sess, hook_post_headers, buf);
+    ne_hook_post_send(sess, hook_post_send, buf);
+    ne_hook_destroy_request(sess, hook_destroy_req, buf);
+    ne_hook_destroy_session(sess, hook_destroy_sess, buf);
+    ne_hook_close_conn(sess, hook_close_conn, buf);
+
+    CALL(any_2xx_request(sess, "/first"));
+
+    ONCMP("(create,GET,/first)\n"
+          "(pre-send)\n"
+          "(post-headers,HTTP/1.1,200,OK)\n"
+          "(post-send,HTTP/1.1,200,OK)\n"
+          "(destroy-req)\n", buf->data, "hook ordering", "first result");
+
+    ne_buffer_clear(buf);
+
+    /* Unhook for mismatched fn/ud pointers: */
+    ne_unhook_create_request(sess, hk_createreq, buf);
+    ne_unhook_create_request(sess, thook_create_req, sess);
+
+    /* Unhook real functions. */
+    ne_unhook_pre_send(sess, hook_pre_send, buf);
+    ne_unhook_destroy_request(sess, hook_destroy_req, buf);
+    ne_unhook_post_headers(sess, hook_post_headers, buf);
+
+    CALL(any_2xx_request(sess, "/second"));
+
+    ONCMP("(create,GET,/second)\n"
+          "(post-send,HTTP/1.1,200,OK)\n", 
+          buf->data, "hook ordering", "second result");
+
+    ne_buffer_clear(buf);
+
+    /* Double hook create, double hook then double unhook post. */
+    ne_hook_create_request(sess, thook_create_req, buf);
+    ne_hook_post_send(sess, hook_post_send, buf);
+    ne_unhook_post_send(sess, hook_post_send, buf);
+    ne_unhook_post_send(sess, hook_post_send, buf);
+
+    CALL(any_2xx_request(sess, "/third"));
+
+    ONCMP("(create,GET,/third)\n"
+          "(create,GET,/third)\n", 
+          buf->data, "hook ordering", "third result");
+
+    ne_buffer_clear(buf);
+
+    ne_session_destroy(sess);
+    CALL(await_server());
+
+    ONCMP("(destroy-sess)\n"
+          "(close-conn)\n", buf->data, "hook ordering", "first destroyed session");
+
+    ne_buffer_clear(buf);
+
+    sess = ne_session_create("http", "www.example.com", 80);
+    ne_hook_destroy_session(sess, hook_destroy_sess, buf);
+    ne_unhook_destroy_session(sess, hook_destroy_sess, buf);
+    ne_session_destroy(sess);
+
+    ONCMP("", buf->data, "hook ordering", "second destroyed session");
+
+    ne_buffer_destroy(buf);
+
+    return OK;
+}
+
+static void hook_self_destroy_req(ne_request *req, void *userdata)
+{
+    ne_unhook_destroy_request(ne_get_session(req), 
+                              hook_self_destroy_req, userdata);
+}
+
+/* Test that it's safe to call ne_unhook_destroy_request from a
+ * destroy_request hook. */
+static int hook_self_destroy(void)
+{
+    ne_session *sess = ne_session_create("http", "localhost", 1234);
+    
+    ne_hook_destroy_request(sess, hook_self_destroy_req, NULL);
+
+    ne_request_destroy(ne_request_create(sess, "GET", "/"));
+
+    ne_session_destroy(sess);
+
+    return OK;
+}
+
+static int icy_protocol(void)
+{
+    ne_session *sess;
+    
+    CALL(make_session(&sess, single_serve_string,
+                      "ICY 200 OK\r\n"
+                      "Content-Length: 0\r\n\r\n"));
+
+    ne_set_session_flag(sess, NE_SESSFLAG_ICYPROTO, 1);
+    
+    ONREQ(any_request(sess, "/foo"));
+
+    ne_session_destroy(sess);
+
+    return await_server();
+}
+
+static void status_cb(void *userdata, ne_session_status status,
+                      const ne_session_status_info *info)
+{
+    ne_buffer *buf = userdata;
+    char scratch[512];
+
+    switch (status) {
+    case ne_status_lookup:
+        ne_buffer_concat(buf, "lookup(", info->lu.hostname, ")-", NULL);
+        break;
+    case ne_status_connecting:
+        ne_iaddr_print(info->ci.address, scratch, sizeof scratch);
+        ne_buffer_concat(buf, "connecting(", info->lu.hostname,
+                         ",", scratch, ")-", NULL);
+        break;
+    case ne_status_disconnected:
+        ne_buffer_czappend(buf, "dis");
+        /* fallthrough */
+    case ne_status_connected:
+        ne_buffer_concat(buf, "connected(", info->cd.hostname, 
+                         ")-", NULL);
+        break;
+    case ne_status_sending:
+    case ne_status_recving:
+        ne_snprintf(scratch, sizeof scratch, 
+                    "%" NE_FMT_NE_OFF_T ",%" NE_FMT_NE_OFF_T, 
+                    info->sr.progress, info->sr.total);
+        ne_buffer_concat(buf, 
+                         status == ne_status_sending ? "send" : "recv",
+                         "(", scratch, ")-", NULL);
+        break;
+    default:
+        ne_buffer_czappend(buf, "bork!");
+        break;
+    }
+}
+
+static int status(void)
+{
+    ne_session *sess;
+    ne_buffer *buf = ne_buffer_create();
+
+    CALL(make_session(&sess, single_serve_string, RESP200
+                      "Content-Length: 5\r\n\r\n" "abcde"));
+
+    ne_set_notifier(sess, status_cb, buf);
+
+    CALL(any_2xx_request_body(sess, "/status"));
+
+    ne_session_destroy(sess);
+    CALL(await_server());
+
+    ONCMP("lookup(localhost)-"
+          "connecting(localhost,127.0.0.1)-"
+          "connected(localhost)-"
+          "send(0,5000)-"
+          "send(5000,5000)-"
+          "recv(0,5)-"
+          "recv(5,5)-"
+          "disconnected(localhost)-",
+          buf->data, "status events", "result");
+    
+    ne_buffer_destroy(buf);
+    
+    return OK;
+}
+
+static int status_chunked(void)
+{
+    ne_session *sess;
+    ne_buffer *buf = ne_buffer_create();
+
+    CALL(make_session(&sess, single_serve_string, 
+                      RESP200 TE_CHUNKED "\r\n" ABCDE_CHUNKS));
+
+    ne_set_notifier(sess, status_cb, buf);
+
+    CALL(any_2xx_request_body(sess, "/status"));
+
+    ne_session_destroy(sess);
+    CALL(await_server());
+    
+    /* This sequence is not guaranteed by the API, but it's what the
+     * current implementation should do. */
+    ONCMP("lookup(localhost)-"
+          "connecting(localhost,127.0.0.1)-"
+          "connected(localhost)-"
+          "send(0,5000)-"
+          "send(5000,5000)-"
+          "recv(0,-1)-"
+          "recv(1,-1)-"
+          "recv(2,-1)-"
+          "recv(3,-1)-"
+          "recv(4,-1)-"
+          "recv(5,-1)-"
+          "disconnected(localhost)-",
+          buf->data, "status events", "result");
+
+    ne_buffer_destroy(buf);
+        
+    return OK;
+}
+
+static const unsigned char raw_127[4] = "\x7f\0\0\01"; /* 127.0.0.1 */
+
+static int local_addr(void)
+{
+    ne_session *sess;
+    ne_inet_addr *ia = ne_iaddr_make(ne_iaddr_ipv4, raw_127);
+
+    CALL(make_session(&sess, single_serve_string, RESP200 
+                      "Connection: close\r\n\r\n"));
+
+    ne_set_localaddr(sess, ia);
+
+    ONREQ(any_request(sess, "/foo"));
+
+    ne_session_destroy(sess);
+    ne_iaddr_free(ia);
+
+    return reap_server();
+}
+
 ne_test tests[] = {
     T(lookup_localhost),
     T(single_get_clength),
@@ -1587,7 +2112,7 @@ ne_test tests[] = {
     T(no_headers),
     T(chunks),
     T(te_header),
-    T(any_te_header),
+    T(te_identity),
     T(reason_phrase),
     T(chunk_numeric),
     T(chunk_extensions),
@@ -1610,6 +2135,7 @@ ne_test tests[] = {
     T(fold_headers),
     T(fold_many_headers),
     T(multi_header),
+    T(multi_header2),
     T(empty_header),
     T(trailing_header),
     T(ignore_header_case),
@@ -1617,12 +2143,18 @@ ne_test tests[] = {
     T(ignore_header_ws2),
     T(ignore_header_ws3),
     T(ignore_header_tabs),
+    T(strip_http10_connhdr),
+    T(strip_http10_connhdr2),
     T(continued_header),
+    T(reset_headers),
+    T(iterate_none),
+    T(iterate_many),
     T(skip_interim_1xx),
     T(skip_many_1xx),
     T(skip_1xx_hdrs),
     T(send_bodies),
     T(expect_100_once),
+    T(expect_100_nobody),
     T(unbounded_headers),
     T(unbounded_folding),
     T(blank_response),
@@ -1632,7 +2164,7 @@ ne_test tests[] = {
     T(fail_eof_chunk),
     T(fail_eof_badclen),
     T(fail_long_header),
-    T(fail_corrupt_chunks),
+    T(fail_on_invalid),
     T(read_timeout),
     T(fail_lookup),
     T(fail_double_lookup),
@@ -1645,5 +2177,13 @@ ne_test tests[] = {
     T(dup_method),
     T(versions),
     T(hook_create_req),
+    T(abort_reader),
+    T(send_bad_offset),
+    T(hooks),
+    T(hook_self_destroy),
+    T(icy_protocol),
+    T(status),
+    T(status_chunked),
+    T(local_addr),
     T(NULL)
 };
